//
//  Copyright (c) 2003-2006 by Autodesk, Inc.
//
//  By using this code, you are agreeing to the terms and conditions of
//  the License Agreement included in the documentation for this code.
//
//  AUTODESK MAKES NO WARRANTIES, EXPRESS OR IMPLIED,
//  AS TO THE CORRECTNESS OF THIS CODE OR ANY DERIVATIVE
//  WORKS WHICH INCORPORATE IT.
//
//  AUTODESK PROVIDES THE CODE ON AN "AS-IS" BASIS
//  AND EXPLICITLY DISCLAIMS ANY LIABILITY, INCLUDING
//  CONSEQUENTIAL AND INCIDENTAL DAMAGES FOR ERRORS,
//  OMISSIONS, AND OTHER PROBLEMS IN THE CODE.
//
//  Use, duplication, or disclosure by the U.S. Government is subject to
//  restrictions set forth in FAR 52.227-19 (Commercial Computer Software
//  Restricted Rights) and DFAR 252.227-7013(c)(1)(ii) (Rights in Technical
//  Data and Computer Software), as applicable.
//

#include "dwfcore/Core.h"

#ifdef  DWFCORE_USE_ANSI_FILE

#include "dwfcore/Pointer.h"
#include "dwfcore/ansi/StandardFileDescriptor.h"
using namespace DWFCore;


_DWFCORE_API
DWFStandardFileDescriptor::DWFStandardFileDescriptor( DWFFile&  rFile,
                                                      int       eIOFlags,
                                                      int       eIOMode )
throw()
                         : DWFFileDescriptor( rFile )
                         , _iHandle( -1 )
                         , _eIOFlags( eIOFlags )
                         , _eIOMode( eIOMode )
{
  OD_TRACE("[DWFExport] DWFStandardFileDescriptor::DWFStandardFileDescriptor\n");
}

_DWFCORE_API
DWFStandardFileDescriptor::DWFStandardFileDescriptor( DWFFile&  rFile,
                                                      int       iHandle )
throw()
                         : DWFFileDescriptor( rFile )
                         , _iHandle( iHandle )
                         , _eIOFlags( 0 )
                         , _eIOMode( 0 )
{
#ifndef NDEBUG
  if (_iHandle > -1)
  { OD_TRACE2("[DWFExport] DWFStandardFileDescriptor::DWFStandardFileDescriptor with open file=%s\n", (const wchar_t*)_oFile.name(), _oFile.name().bytes()); }
  else
  { OD_TRACE("[DWFExport] DWFStandardFileDescriptor::DWFStandardFileDescriptor with open\n"); }
#endif
}

_DWFCORE_API
DWFStandardFileDescriptor::~DWFStandardFileDescriptor()
throw()
{
  OD_TRACE("[DWFExport] DWFStandardFileDescriptor::~DWFStandardFileDescriptor\n");
  try { close(); } catch (...)
  { OD_TRACE("[DWFExport] DWFStandardFileDescriptor::~DWFStandardFileDescriptor failed\n"); }
}

_DWFCORE_API
void
DWFStandardFileDescriptor::open()

{
        //
        // already open
        //
    if (_iHandle > -1)
    {
        OD_TRACE("[DWFExport] DWFStandardFileDescriptor::open already\n");
        _DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"File already open" );
    }
    OD_TRACE2("[DWFExport] DWFStandardFileDescriptor::open file=%s\n", (const wchar_t*)_oFile.name(), _oFile.name().bytes());

    //
    // always use UTF-8 encoded filenames
    //
    char* pUTF8Filename = NULL;
    const DWFString& rFilename = _oFile.name();

    //
    // get the encoded filename
    //
    rFilename.getUTF8( &pUTF8Filename );

    //
    // open the file per the current option flags
    //
    _iHandle = ::open( pUTF8Filename, _eIOFlags, _eIOMode );

    //
    // free memory
    //
    DWFCORE_FREE_MEMORY( pUTF8Filename );

        //
        // check stream
        //
    if (_iHandle == -1)
    {
      OD_TRACE("[DWFExport] DWFStandardFileDescriptor::open failed\n");
        _DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"Failed to open file" );
    }
}

_DWFCORE_API
void
DWFStandardFileDescriptor::close()

{
    OD_TRACE("[DWFExport] DWFStandardFileDescriptor::close\n");
        //
        // let's be forgiving and not throw exceptions
        //
    if (_iHandle > -1)
    {
        ::close( _iHandle );

        _iHandle = -1;
        OD_TRACE("[DWFExport] DWFStandardFileDescriptor::close ok\n");
    }
}

_DWFCORE_API
size_t
DWFStandardFileDescriptor::size()

{
    OD_TRACE("[DWFExport] DWFStandardFileDescriptor::size\n");
    if (_iHandle == -1)
    {
        OD_TRACE("[DWFExport] DWFStandardFileDescriptor::size filed\n");
        _DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"No file currently open" );
    }

    struct stat tStat;

    if (::fstat(_iHandle, &tStat) == -1)
    {
        OD_TRACE("[DWFExport] DWFStandardFileDescriptor::size filed 1\n");
        _DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"Failed to stat file" );
    }

    OD_TRACE1("[DWFExport] DWFStandardFileDescriptor::size=%d\n", (void*)tStat.st_size);

    return tStat.st_size;
}

_DWFCORE_API
off_t
DWFStandardFileDescriptor::seek( int eOrigin,
                                 off_t              nOffset )

{
    if (_iHandle == -1)
    {
        _DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"No file currently open" );
    }

    if (::lseek(_iHandle, nOffset, eOrigin) == -1)
    {
        _DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"Failed to seek" );
    }

    //
    // not meaningful
    //
    return -1;
}

_DWFCORE_API
size_t
DWFStandardFileDescriptor::read( void* pBuffer, size_t nBufferBytes )
{
    if (_iHandle == -1)
    {
        OD_TRACE("[DWFExport] DWFStandardFileDescriptor::read failed\n");
        _DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"No file currently open" );
    }

    OD_TRACE1("[DWFExport] DWFStandardFileDescriptor::read requested=%d\n", (void*)nBufferBytes);
    size_t nBytesRead = ::read( _iHandle, pBuffer, nBufferBytes );
    if (nBytesRead < 0)
    {
        OD_TRACE("[DWFExport] DWFStandardFileDescriptor::read failed 1\n");
        _DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"Read failed" );
    }
    OD_TRACE1("[DWFExport] DWFStandardFileDescriptor::read=%d :\n", (void*)nBytesRead);
    OD_TRACE1("    (0x%lX 0x%lX 0x%lX 0x%lX 0x%lX ...)\n", pBuffer);

    return nBytesRead;
}

_DWFCORE_API
size_t
DWFStandardFileDescriptor::write( const void* pBuffer, size_t nBufferBytes )

{
    if (_iHandle == -1)
    {
        OD_TRACE("[DWFExport] DWFStandardFileDescriptor::write failed\n");
        _DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"No file currently open" );
    }

    OD_TRACE1("[DWFExport] DWFStandardFileDescriptor::write requested=%d\n", (void*)nBufferBytes);
    size_t nBytesWritten = ::write( _iHandle, pBuffer, nBufferBytes );
    if (nBytesWritten != nBufferBytes) //< 0)
    {
        OD_TRACE("[DWFExport] DWFStandardFileDescriptor::write failed 1\n");
        _DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"Write failed" );
    }
    OD_TRACE1("[DWFExport] DWFStandardFileDescriptor::written=%d\n", (void*)nBytesWritten);

    return nBytesWritten;
}

#endif
