#include "dwfcore/paging/SqlAdapter.h"
#include "dwfcore/Exception.h"
#include <cassert>

namespace DWFCore
{
	DWFPagingAdapter* default_adapter = 0;

  DWFSqliteAdapter::DWFSqliteAdapter()
    : _pBTree(0)
    , _pCur(0)
  {
    // SUP-5571 // _open();
#ifdef _DEBUG
    if (false)
      _open(); // test to check with debugger on android
#endif
  }
  void DWFSqliteAdapter::_open()
  {
    if (_pBTree)
      return;
    //
#ifdef __IPHONE_OS_VERSION_MIN_REQUIRED
    static char sTempDir[1024];
    size_t nSize = confstr(_CS_DARWIN_USER_TEMP_DIR, NULL, 0);
    size_t	 size = confstr(_CS_DARWIN_USER_TEMP_DIR, sTempDir, nSize);// __DARWIN_ALIAS(confstr);
    sqlite3_temp_directory =  sTempDir;
#endif
#ifdef ANDROID
    /* AMark : typically all Android devices contains /sdcard mounted. If external card doesn't connected, internal device
               storage is mounted on this point. Local application temporary path is available from Java layer only, so
               we can't access it here. Probably better to use drawing path instead, but it is also could be protected.
               So, the best way to find temporary directory on Android is a still opened for discussion.                   */
    static char sTempDir[] = "/sdcard"; // sqlite3_temp_directory isn't const, so store sTempDir non-const
    if (!sqlite3_temp_directory) // Set only if doesn't set from app
      sqlite3_temp_directory = sTempDir;
#endif
		int rc = sqlite3BtreeOpen(0, 0, &_pBTree, BTREE_OMIT_JOURNAL | BTREE_NO_READLOCK);

		if (SQLITE_OK == rc)
		{
			tScopedTransaction oScopedTrans(*this);
			int nTableId;
			rc = sqlite3BtreeCreateTable(_pBTree, &nTableId, BTREE_INTKEY|BTREE_LEAFDATA );
			rc = sqlite3BtreeCursor(_pBTree, nTableId, 1, NULL, 0, &_pCur);
			oScopedTrans.setSuccess();
		}

		if (SQLITE_OK != rc)
		{
      if (_pBTree != 0)
      {
        sqlite3BtreeClose(_pBTree);
        _pBTree = NULL;
      }

			_DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"Open database failed");
		}
	}

  DWFSqliteAdapter::~DWFSqliteAdapter()
  {
    _close();
  }

  void DWFSqliteAdapter::_close()
	{
    if (_pCur)
    {
      sqlite3BtreeCloseCursor(_pCur);
      _pCur = NULL;
    }
    if (_pBTree)
    {
      sqlite3BtreeClose(_pBTree);
      _pBTree = NULL;
    }
	}

	void DWFSqliteAdapter::load(DWFSqliteAdapter::tHandleType hHandle, DWFSqliteAdapter::tBufferType& buf)
	{
    _open();

		int res;
		int rc = sqlite3BtreeMoveto(_pCur, 0, (i64)hHandle, true, &res);
		if (SQLITE_OK == rc)
		{
			assert(res == 0);
			u32 size;
			sqlite3BtreeDataSize(_pCur, &size);
			buf.resize(size);
			sqlite3BtreeData(_pCur, 0, size, &buf[0]);
		}
		else
		{
			_DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"Move cursor failed");
		}		
	}

	DWFSqliteAdapter::tHandleType DWFSqliteAdapter::save(const DWFSqliteAdapter::tBufferType& src, DWFSqliteAdapter::tHandleType old)
	{
		static tHandleType _nextHandle = 1;
   _open(); // SUP-5571

		tHandleType  new_handle = old == 0 ? _nextHandle++ : old;

		int rc = sqlite3BtreeInsert(_pCur, 0, (i64)new_handle, &src[0], (int)src.size(), old == 0);
		if (SQLITE_OK != rc)
		{
			_DWFCORE_THROW( DWFIOException, /*NOXLATE*/L"Insertion failed");
		}
		return new_handle;
	}

	void DWFSqliteAdapter::begin()
	{
    _open(); // SUP-5571
    sqlite3BtreeBeginTrans(_pBTree, 1);
	}

	void DWFSqliteAdapter::commit()
	{
    _open(); // SUP-5571
    sqlite3BtreeCommit(_pBTree);
	}
	void DWFSqliteAdapter::rollback()
	{
    _open(); // SUP-5571
    sqlite3BtreeRollback(_pBTree);
	}
}
